import copy
import itertools
import torch
from torch import nn, Tensor
from typing import Iterable, List, Sequence, Tuple
from .block import FcBody, LinearHead, NetworkBlock
from .init import Initializer, fanin_uniform, constant
from .misc import SoftUpdate
from .policy import Policy, PolicyDist, TanhGaussianDist
from .prelude import NetFn
from .value import ContinuousQFunction
from ..prelude import ArrayLike, Self
from ..utils import Device


class SACTarget(SoftUpdate):
    def __init__(self, critic1: nn.Module, critic2: nn.Module, device: Device) -> None:
        super().__init__()
        self.critic1 = critic1
        self.critic2 = critic2
        self.device = device

    def soft_update(self, other: Self, coef: float) -> None:
        SoftUpdate.soft_update(self.critic1, other.critic1, coef)  # type: ignore
        SoftUpdate.soft_update(self.critic2, other.critic2, coef)  # type: ignore

    def q_values(self, states: ArrayLike, action: ArrayLike) -> Tuple[Tensor, Tensor]:
        sa = torch.cat((self.device.tensor(states), self.device.tensor(action)), dim=1)
        return self.critic1(sa), self.critic2(sa)


class SeparatedSACNet(nn.Module, ContinuousQFunction):
    def __init__(
        self,
        actor_body: NetworkBlock,
        critic_body1: NetworkBlock,
        critic_body2: NetworkBlock,
        policy_dist: PolicyDist,
        device: Device = Device(),
        init: Initializer = Initializer(
            weight_init=fanin_uniform(), bias_init=constant(0.1)
        ),
    ) -> None:
        super().__init__()
        self.actor = nn.Sequential(
            actor_body,
            LinearHead(actor_body.output_dim, policy_dist.input_dim, init=init),
        )
        self.critic1 = nn.Sequential(
            critic_body1, LinearHead(critic_body1.output_dim, 1, init=init)
        )
        self.critic2 = nn.Sequential(
            critic_body2, LinearHead(critic_body2.output_dim, 1, init=init)
        )
        self.policy_dist = policy_dist
        self.device = device
        self.to(device.unwrapped)

    def q_value(self, states: ArrayLike, action: ArrayLike) -> Tensor:
        sa = torch.cat((self.device.tensor(states), self.device.tensor(action)), dim=1)
        return self.critic1(sa)

    def q_values(self, states: ArrayLike, action: ArrayLike) -> Tuple[Tensor, Tensor]:
        sa = torch.cat((self.device.tensor(states), self.device.tensor(action)), dim=1)
        return self.critic1(sa), self.critic2(sa)

    def get_target(self) -> SACTarget:
        return SACTarget(
            copy.deepcopy(self.critic1), copy.deepcopy(self.critic2), self.device
        )

    def policy(self, states: ArrayLike) -> Policy:
        st = self.device.tensor(states)
        if st.dim() == 1:
            st = st.view(1, -1)
        policy_params = self.actor(st)
        return self.policy_dist(policy_params)

    def actor_params(self) -> Iterable[Tensor]:
        return self.actor.parameters()

    def critic_params(self) -> Iterable[Tensor]:
        return itertools.chain(self.critic1.parameters(), self.critic2.parameters())

    def forward(
        self, states: ArrayLike, action: ArrayLike
    ) -> Tuple[Tensor, Tensor, Policy]:
        s, a = self.device.tensor(states), self.device.tensor(action)
        sa = torch.cat((s, a), dim=1)
        q1, q2 = self.critic1(sa), self.critic2(sa)
        policy = self.policy_dist(self.actor(s))
        return q1, q2, policy


def fc_separated(
    actor_units: List[int] = [256, 256],
    critic_units: List[int] = [256, 256],
    policy_type: type = TanhGaussianDist,
    init: Initializer = Initializer(
        weight_init=fanin_uniform(), bias_init=constant(0.1)
    ),
) -> NetFn:
    """SAC network with separated bodys
    """

    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> SeparatedSACNet:
        actor_body = FcBody(state_dim[0], units=actor_units, init=init)
        critic1 = FcBody(state_dim[0] + action_dim, units=critic_units, init=init)
        critic2 = FcBody(state_dim[0] + action_dim, units=critic_units, init=init)
        policy = policy_type(action_dim)
        return SeparatedSACNet(
            actor_body, critic1, critic2, policy, device=device, init=init,
        )

    return _net



import numpy as np
from typing import Callable, Generic, List, NamedTuple, Tuple, Type
from .array_deque import ArrayDeque
from .base import ReplayFeed, ReplayBuffer
from ..utils import sample_indices
from ..prelude import Array, GenericNamedMeta, State


class UniformReplayBuffer(ReplayBuffer, Generic[ReplayFeed]):
    def __init__(self, feed: Type[ReplayFeed], capacity: int = 1000) -> None:
        super().__init__(feed)
        self.buf = ArrayDeque(capacity=capacity)
        self.cap = capacity

    def append(self, *args) -> None:
        exp = self.feed(*args)
        self.buf.push_back(exp)
        if len(self) > self.cap:
            self.buf.pop_front()

    def sample(self, batch_size: int) -> List[ReplayFeed]:
        if self.allow_overlap:
            indices = np.random.randint(len(self.buf), size=batch_size)
        else:
            indices = sample_indices(len(self.buf), batch_size)
        return [self.buf[idx] for idx in indices]

    def __len__(self):
        return len(self.buf)


class DQNReplayFeed(NamedTuple, Generic[State], metaclass=GenericNamedMeta):
    state: State
    action: int
    reward: float
    next_state: State
    done: bool

    def to_array(
        self, wrap: Callable[[State], Array]
    ) -> Tuple[Array[float], int, float, Array[float], bool]:
        return (
            wrap(self.state),
            self.action,
            self.reward,
            wrap(self.next_state),
            self.done,
        )


class BootDQNReplayFeed(NamedTuple, Generic[State], metaclass=GenericNamedMeta):
    state: State
    action: int
    reward: float
    next_state: State
    done: bool
    ensemble_mask: Array[bool]

    def to_array(
        self, wrap: Callable[[State], Array]
    ) -> Tuple[Array[float], int, float, Array[float], bool, Array[bool]]:
        return (
            wrap(self.state),
            self.action,
            self.reward,
            wrap(self.next_state),
            self.done,
            self.ensemble_mask,
        )



from abc import ABC, abstractmethod
import numpy as np
from torch import nn, Tensor
from typing import Sequence, Tuple
from .block import DQNConv, FcBody, LinearHead, NetworkBlock
from .prelude import NetFn
from ..utils import Device
from ..prelude import Array, ArrayLike


class ContinuousQFunction(ABC):
    @abstractmethod
    def q_value(self, states: ArrayLike, action: ArrayLike) -> Tensor:
        pass


class DiscreteQFunction(ABC):
    @abstractmethod
    def q_value(self, state: Array, nostack: bool = False) -> Tensor:
        pass

    @property
    @abstractmethod
    def state_dim(self) -> Sequence[int]:
        pass

    @property
    @abstractmethod
    def action_dim(self) -> int:
        pass


class DiscreteQValueNet(DiscreteQFunction, nn.Module):
    """State -> [Value..]
    """

    def __init__(
        self,
        body: NetworkBlock,
        head: NetworkBlock,
        device: Device = Device(),
        do_not_use_data_parallel: bool = False,
    ) -> None:
        if body.output_dim != np.prod(head.input_dim):
            raise ValueError("body output and head input must have a same dimention")
        super().__init__()
        self.head = head
        self.body = body
        if not do_not_use_data_parallel and device.is_multi_gpu():
            self.body = device.data_parallel(body)  # type: ignore
        self.device = device
        self.to(self.device.unwrapped)

    def q_value(self, state: Array, nostack: bool = False) -> Tensor:
        if nostack:
            return self.forward(state)
        else:
            return self.forward(np.stack([state]))

    def forward(self, x: ArrayLike) -> Tensor:
        x = self.device.tensor(x)
        x = self.body(x)
        x = self.head(x)
        return x

    @property
    def state_dim(self) -> Sequence[int]:
        return self.body.input_dim

    @property
    def action_dim(self) -> int:
        return self.head.output_dim


def dqn_conv(*args, **kwargs) -> NetFn:
    def _net(
        state_dim: Tuple[int, int, int], action_dim: int, device: Device
    ) -> DiscreteQValueNet:
        body = DQNConv(state_dim, *args, **kwargs)
        head = LinearHead(body.output_dim, action_dim)
        return DiscreteQValueNet(body, head, device=device)

    return _net  # type: ignore


def fc(*args, **kwargs) -> NetFn:
    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> DiscreteQValueNet:
        body = FcBody(state_dim[0], *args, **kwargs)
        head = LinearHead(body.output_dim, action_dim)
        return DiscreteQValueNet(body, head, device=device)

    return _net



import gym
from gym.utils import seeding
import numpy as np
from typing import List, Optional, Tuple, Union


class DeepSea(gym.Env):
    SCREEN_SIZE = 400

    def __init__(self, size: int, noise: float = 0.0) -> None:
        self._size = size
        self._move_cost = 0.01 / size
        self._goal_reward = 1.0
        self._column = 0
        self._row = 0
        self.action_space = gym.spaces.Discrete(2)
        low = np.zeros(size ** 2)
        high = np.ones(size ** 2)
        self.observation_space = gym.spaces.Box(low, high)
        self.np_random = None
        self.noise = noise
        self._viewer = None

    def seed(self, seed: Optional[int] = None) -> List[int]:
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, action: int) -> Tuple[np.ndarray, float, bool, dict]:
        # Remap actions according to column (action_right = go right)
        if self.noise == 0.0 or self.noise < self.np_random.uniform(0, 1):
            action_right = action == 1
        else:
            action_right = action != 1

        # Compute the reward
        reward = 0.0
        if self._column == self._size - 1 and action_right:
            reward += self._goal_reward

        # State dynamics
        if action_right:  # right
            self._column = np.clip(self._column + 1, 0, self._size - 1)
            reward -= self._move_cost
        else:  # left
            self._column = np.clip(self._column - 1, 0, self._size - 1)

        # Compute the observation
        self._row += 1
        if self._row == self._size:
            observation = self._get_observation(self._row - 1, self._column)
            return observation, reward, True, {}
        else:
            observation = self._get_observation(self._row, self._column)
            return observation, reward, False, {}

    def reset(self) -> np.ndarray:
        self._reset_next_step = False
        self._column = 0
        self._row = 0
        return self._get_observation(self._row, self._column)

    def render(self, mode: str = "human") -> Union[np.ndarray, bool]:
        player_size = self.SCREEN_SIZE / self._size
        if self._viewer is None:
            from gym.envs.classic_control import rendering

            self._viewer = rendering.Viewer(self.SCREEN_SIZE, self.SCREEN_SIZE)
            self.player_trans = rendering.Transform()
            v = np.array([(0.0, 0.0), (2.0, 0.0), (1.5, -1.0), (0.5, -1.0)])
            player = rendering.make_polygon(v * player_size / 2)
            player.set_color(0.0, 0.0, 1.0)
            player.add_attr(self.player_trans)
            self._viewer.add_geom(player)
        x = player_size * self._column
        y = self.SCREEN_SIZE - player_size * self._row
        self.player_trans.set_translation(x, y)
        return self._viewer.render(return_rgb_array=mode == "rgb_array")

    def _get_observation(self, row: int, column: int) -> np.ndarray:
        observation = np.zeros(shape=(self._size, self._size), dtype=np.float32)
        observation[row, column] = 1
        return observation.flatten()



import torch
from ..prelude import Self


class SoftUpdate(torch.nn.Module):
    @torch.no_grad()
    def soft_update(self, other: Self, coef: float) -> None:
        for s_param, o_param in zip(self.parameters(), other.parameters()):
            s_param.copy_(s_param * (1.0 - coef) + o_param * coef)



import atexit
import click
from collections import defaultdict
import datetime as dt
import pandas as pd
from pandas import DataFrame
from pathlib import Path
from typing import Any, DefaultDict, Dict, List, NamedTuple, Tuple, Union
import warnings


class LogStore:
    """A temporal object for storing logs.
    Since Pandas DataFrame/Series's append is not efficient, we store logs
    in this object before converting it into Pandas objects.
    """

    def __init__(self) -> None:
        self.inner: DefaultDict[str, List[Any]] = defaultdict(list)

    def submit(self, d: Dict[str, Any]) -> int:
        res = 0
        for key, value in d.items():
            self.inner[key].append(value)
            if res == 0:
                res = len(self.inner[key])
        return res

    def into_df(self) -> DataFrame:
        df = DataFrame(self.inner)
        self.inner.clear()
        return df

    def to_df(self) -> DataFrame:
        return DataFrame(self.inner)

    def reset(self) -> None:
        self.inner.clear()

    def __getitem__(self, index: Union[int, slice]) -> "LogStore":
        res: DefaultDict[str, List[Any]] = defaultdict(list)
        for key, value in self.inner.items():
            res[key] = value[index]
        log_store = LogStore()
        log_store.inner = res
        return log_store

    def __len__(self) -> int:
        return len(self.inner)

    def __repr__(self) -> str:
        return "LogStore({})".format(dict.__repr__(self.inner))


class SummarySetting(NamedTuple):
    indices: List[str]
    interval: int
    color: str


class ExperimentLogger:
    """
    - stores experiment log
    - exposes logs as pandas.DataFrame
    - prints summaries of logs to stdout
    """

    FINGERPRINT = "fingerprint.txt"
    LOG_CAPACITY = int(1e6)

    def __init__(self, show_summary: bool = True) -> None:
        self.logdir = Path("Results/Temp")
        self.exp_start = dt.datetime.now()
        self.exp_name = "No name"

        self._store: DefaultDict[str, LogStore] = defaultdict(LogStore)
        self._summary_setting: DefaultDict[str, SummarySetting] = defaultdict(
            lambda: SummarySetting(["time"], 1, "black")
        )
        self._show_summary = show_summary
        self._closed = False
        self._time_offset = dt.timedelta(0)
        self.ready = False
        atexit.register(self.close)

    def setup_from_script_path(
        self, script_path_: str, prefix: str = "", fingerprint: Dict[str, str] = {},
    ) -> None:
        script_path = Path(script_path_)
        logdir_top = "Results"
        if prefix:
            logdir_top = prefix + "/" + logdir_top
        self.exp_name = script_path.stem
        time = self.exp_start.strftime("%y%m%d-%H%M%S")
        self.logdir = Path(logdir_top).joinpath(self.exp_name).joinpath(time)
        fingerprint.update(self.git_metadata(script_path))
        self.setup_logdir(fingerprint=fingerprint)

    def retrive(self, logdir: Path) -> Tuple[int, int]:
        self.logdir = logdir
        sec, total_steps, episodes = 0, 0, 0
        for csvlog in self.logdir.glob("*.csv"):
            df = pd.read_csv(csvlog.as_posix())
            last = df.iloc[-1]
            sec = max(sec, last["sec"])
            if "total_steps" in last:
                total_steps = max(total_steps, last["total_steps"])
            if "episodes" in last:
                episodes = max(episodes, last["episodes"])
        self._time_offset = dt.timedelta(seconds=sec)
        return total_steps, episodes

    @staticmethod
    def git_metadata(script_path: Path) -> dict:
        res: Dict["str", "str"] = {}
        try:
            import git
        except ImportError:
            warnings.warn("GitPython is not Installed")
            return res
        try:
            repo = git.Repo(script_path, search_parent_directories=True)
            head = repo.head.commit
            res["git-head"] = head.hexsha
            res["git-diff"] = repo.git.diff()
        except git.exc.InvalidGitRepositoryError:
            warnings.warn("{} is not in a git repository".format(script_path))
        return res

    def setup_logdir(self, fingerprint: Dict[str, str] = {}) -> None:
        if not self.logdir.exists():
            self.logdir.mkdir(parents=True)
        finger = self.logdir.joinpath(self.FINGERPRINT)
        with finger.open(mode="w") as f:
            f.write("name: {}\nstarttime: {}\n".format(self.exp_name, self.exp_start))
            for fkey in fingerprint.keys():
                f.write("{}: {}\n".format(fkey, fingerprint[fkey]))
        self.ready = True

    def submit(self, name: str, **kwargs) -> None:
        """Stores log.
        """
        kwargs["sec"] = (dt.datetime.now() - self.exp_start).total_seconds()
        current_length = self._store[name].submit(kwargs)
        if self._show_summary:
            interval = self._summary_setting[name].interval
            if current_length % interval == 0:
                self.show_summary(name)
        if current_length >= self.LOG_CAPACITY:
            self._truncate_and_dump(name)

    def summary_setting(
        self, name: str, indices: List[str], interval: int = 1, color: str = "black",
    ) -> None:
        if "sec" not in indices:
            indices.append("sec")
        self._summary_setting[name] = SummarySetting(indices, interval, color)

    def show_summary(self, name: str) -> None:
        indices, interval, color = self._summary_setting[name]
        df = self._store[name][-interval:].to_df()
        indices_df = df[indices]
        click.secho(
            f"=========={name.upper()} LOG===========", bg=color, fg="white", bold=True
        )
        min_, max_ = indices_df.iloc[0], indices_df.iloc[-1]
        range_str = "\n".join([f"{idx}: {min_[idx]}-{max_[idx]}" for idx in indices])
        click.secho(range_str, bg="black", fg="white")
        df.drop(columns=indices, inplace=True)
        describe = df.describe()
        describe.drop(labels="count", inplace=True)
        click.echo(df.describe())

    def close(self) -> None:
        if self._closed:
            return
        for key in self._store.keys():
            self._truncate_and_dump(key)
        self._closed = True

    def _truncate_and_dump(self, name: str) -> None:
        df = self._store[name].into_df()
        path = self.logdir.joinpath(name + ".csv")
        include_header = not path.exists()
        mode = "w" if include_header else "a"
        df.to_csv(path.as_posix(), mode=mode, header=include_header, index=False)



from abc import abstractmethod
from copy import deepcopy
import numpy as np
import torch
from torch import nn, Tensor
from torch.nn import functional as F
from typing import List, Sequence
from .block import FcBody, LinearHead, NetworkBlock
from .init import Initializer, xavier_uniform
from .prelude import NetFn
from .value import DiscreteQFunction, DiscreteQValueNet
from ..utils import Device
from ..prelude import Array, ArrayLike


class BootstrappedQFunction(DiscreteQFunction):
    device: Device

    @abstractmethod
    def forward(self, states: ArrayLike) -> Tensor:
        pass

    @abstractmethod
    def q_i_s(self, index: int, states: ArrayLike) -> Tensor:
        pass

    def q_value(self, state: Array) -> Tensor:
        values = self(state)
        return values.mean(dim=1)

    def q_s_a(self, states: ArrayLike, actions: ArrayLike) -> Tensor:
        qs = self(self.device.tensor(states))
        act = self.device.tensor(actions, dtype=torch.long)
        action_mask = F.one_hot(act, num_classes=qs.size(-1)).float()
        return torch.einsum("bka,ba->bk", qs, action_mask)


class SeparatedBootQValueNet(BootstrappedQFunction, nn.Module):
    def __init__(self, q_nets: List[DiscreteQFunction]):
        super().__init__()
        self.q_nets = nn.ModuleList(q_nets)
        self.device = q_nets[0].device

    def forward(self, x: ArrayLike) -> Tensor:
        return torch.stack([q(x) for q in self.q_nets], dim=1)

    def q_i_s(self, index: int, states: ArrayLike) -> Tensor:
        return self.q_nets[index](states)

    @property
    def state_dim(self) -> Sequence[int]:
        return self.q_nets[0].state_dim

    @property
    def action_dim(self) -> int:
        return self.q_nets[0].action_dim


class SharedBootQValueNet(DiscreteQFunction, nn.Module):
    """INCOMPLETE
    """

    def __init__(
        self, body: NetworkBlock, heads: List[NetworkBlock], device: Device = Device()
    ) -> None:
        if body.output_dim != np.prod(head[0].input_dim):
            raise ValueError("body output and head input must have a same dimention")
        super().__init__()
        self.body = body
        self.head = nn.ModuleList(head)
        self.device = device
        self.to(self.device.unwrapped)

    def forward(self, index: int, x: ArrayLike) -> Tensor:
        raise NotImplementedError()

    @property
    def state_dim(self) -> Sequence[int]:
        return self.body.input_dim

    @property
    def action_dim(self) -> int:
        return self.head[0].output_dim


class PriorQValueNet(DiscreteQFunction, nn.Module):
    """State -> [Value..]
    """

    def __init__(
        self,
        body: NetworkBlock,
        head: NetworkBlock,
        prior_scale: float = 1.0,
        device: Device = Device(),
        init: Initializer = Initializer(),
    ) -> None:
        if body.output_dim != np.prod(head.input_dim):
            raise ValueError("body output and head input must have a same dimention")
        super().__init__()
        self.model = nn.Sequential(body, head)
        self.prior = init(deepcopy(self.model))
        self.device = device
        self.prior_scale = prior_scale
        self.to(self.device.unwrapped)

    def q_value(self, state: Array, nostack: bool = False) -> Tensor:
        if nostack:
            return self.forward(state)
        else:
            return self.forward(np.stack([state]))

    def forward(self, x: ArrayLike) -> Tensor:
        x = self.device.tensor(x)
        raw = self.model(x)
        with torch.no_grad():
            prior = self.prior(x)
        return raw.add_(prior.mul_(self.prior_scale))

    @property
    def state_dim(self) -> Sequence[int]:
        return self.model[0].input_dim

    @property
    def action_dim(self) -> int:
        return self.model[1].output_dim


def fc_separated(n_ensembles: int, *args, **kwargs) -> NetFn:
    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> SeparatedBootQValueNet:
        q_nets = []
        for _ in range(n_ensembles):
            body = FcBody(state_dim[0], *args, **kwargs)
            head = LinearHead(body.output_dim, action_dim)
            q_nets.append(DiscreteQValueNet(body, head, device=device))
        return SeparatedBootQValueNet(q_nets)

    return _net


def rpf_fc_separated(
    n_ensembles: int,
    prior_scale: float = 1.0,
    init: Initializer = Initializer(weight_init=xavier_uniform()),
    **kwargs
) -> NetFn:
    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> SeparatedBootQValueNet:
        q_nets = []
        for _ in range(n_ensembles):
            body = FcBody(state_dim[0], init=init, **kwargs)
            head = LinearHead(body.output_dim, action_dim, init=init)
            prior_model = PriorQValueNet(body, head, prior_scale, init=init)
            q_nets.append(prior_model)
        return SeparatedBootQValueNet(q_nets)

    return _net



import torch
from torch import Tensor


@torch.jit.script
def clamp_actions_(t: Tensor, min_t: Tensor, max_t: Tensor) -> Tensor:
    for i in range(t.size(1)):
        t[:, i].clamp_(min_t[i].item(), max_t[i].item())
    return t


@torch.jit.script
def normalize_(t: Tensor, eps: float) -> None:
    mean = t.mean()
    std = t.std()
    t.sub_(mean).div_(std + eps)



import click
from typing import Callable, List, Optional

from ..agents import Agent
from ..config import Config
from ..experiment import Experiment
from ..lib import mpi


@click.group()
@click.option(
    "--envname", type=str, default=None, help="Name of environment passed to config_gen"
)
@click.option("--max-steps", type=int, default=None, help="Max steps of the training")
@click.option(
    "--seed",
    type=int,
    default=None,
    help="Random seed set before training. Left for backward comaptibility",
)
@click.option("--model", type=str, default=None, help="Name of the save file")
@click.option(
    "--action-file", type=str, default="actions.json", help="Name of the action file",
)
@click.pass_context
def rainy_cli(
    ctx: click.Context,
    envname: Optional[str],
    max_steps: Optional[int],
    seed: Optional[int],
    model: Optional[str],
    action_file: Optional[str],
    **kwargs,
) -> None:
    cfg_gen = ctx.obj["config_gen"]
    if envname is not None:
        kwargs["envname"] = envname
    if max_steps is not None:
        kwargs["max_steps"] = max_steps
    config = cfg_gen(**kwargs)
    config.seed = seed
    ag = ctx.obj["agent_gen"](config)
    ctx.obj["experiment"] = Experiment(ag, model, action_file)
    ctx.obj["envname"] = "Default" if envname is None else envname
    ctx.obj["kwargs"] = kwargs


@rainy_cli.command(help="Train agents")
@click.pass_context
@click.option(
    "--comment",
    type=str,
    default=None,
    help="Comment that would be wrote to fingerprint.txt",
)
@click.option("--prefix", type=str, default="", help="Prefix of the log directory")
@click.option(
    "--eval-render", is_flag=True, help="Render the environment when evaluating"
)
def train(
    ctx: click.Context, comment: Optional[str], prefix: str, eval_render: bool = True
) -> None:
    script_path = ctx.obj["script_path"]
    experiment = ctx.obj["experiment"]
    if script_path is not None:
        fingerprint = dict(
            comment="" if comment is None else comment,
            envname=ctx.obj["envname"],
            kwargs=ctx.obj["kwargs"],
        )
        experiment.logger.setup_from_script_path(
            script_path, prefix=prefix, fingerprint=fingerprint
        )
    experiment.train(eval_render=eval_render)
    if mpi.IS_MPI_ROOT:
        print(
            "random play: {}, trained: {}".format(
                experiment.ag.random_episode(), experiment.ag.eval_episode()
            )
        )


@rainy_cli.command(help="Run the random agent and show its result")
@click.option("--save", is_flag=True, help="Save actions")
@click.option("--render", is_flag=True, help="Render the agent")
@click.option(
    "--replay",
    is_flag=True,
    help="Show replay(works only with special environments, e.g., rogue-gym)",
)
@click.pass_context
def random(ctx: click.Context, save: bool, render: bool, replay: bool) -> None:
    experiment = ctx.obj["experiment"]
    if save:
        experiment.random(render=render, replay=replay, action_file=action_file)
    else:
        experiment.random(render=render, replay=replay)


@rainy_cli.command(help="Given a save file and restart training")
@click.pass_context
@click.argument("logdir", type=str)
@click.option(
    "--additional-steps",
    type=int,
    default=100,
    help="The number of  additional training steps",
)
@click.option(
    "--eval-render", is_flag=True, help="Render the environment when evaluating"
)
def retrain(
    ctx: click.Context, logdir: str, additional_steps: int, eval_render: bool
) -> None:
    experiment = ctx.obj["experiment"]
    experiment.retrain(logdir, additional_steps, eval_render)
    if mpi.IS_MPI_ROOT:
        print(
            "random play: {}, trained: {}".format(
                experiment.ag.random_episode(), experiment.ag.eval_episode()
            )
        )


@rainy_cli.command(help="Load a specified save file and evaluate the agent")
@click.argument("logdir", type=str)
@click.option("--render", is_flag=True, help="Render the agent")
@click.option(
    "--replay",
    is_flag=True,
    help="Show replay(works only with special environments, e.g., rogue-gym)",
)
@click.pass_context
def eval(ctx: click.Context, logdir: str, render: bool, replay: bool,) -> None:
    experiment = ctx.obj["experiment"]
    experiment.evaluate(logdir, render=render, replay=replay)


def _add_options(options: List[click.Command]) -> click.Group:
    for option in options:
        rainy_cli.params.append(option)


def run_cli(
    config_gen: Callable[..., Config],
    agent_gen: Callable[[Config], Agent],
    script_path: Optional[str] = None,
    options: List[click.Command] = [],
) -> None:
    obj = {
        "config_gen": config_gen,
        "agent_gen": agent_gen,
        "script_path": script_path,
    }
    _add_options(options)
    rainy_cli(obj=obj)



from abc import ABC, abstractmethod
import torch
from torch import nn, Tensor
from torch.distributions import Bernoulli, Categorical, Distribution, Normal
from torch.nn import functional as F
from typing import Optional, Tuple
from ..prelude import Array, Index, Self
from ..utils import Device


class Policy(ABC):
    """Represents parameterized stochastic policies.
    """

    def __init__(self, dist: Distribution) -> None:
        self.dist = dist
        self._action: Optional[Tensor] = None
        self._baction: Optional[Tensor] = None

    def action(self) -> Tensor:
        """Sample actions if this policy has no action cache.
        """
        if self._action is None:
            self._action = self.sample()
        return self._action.squeeze()

    def baction(self) -> Tensor:
        """Sample "backwardable" actions.
        """
        if self._baction is None:
            self._baction = self.rsample()
        return self._baction.squeeze()

    def set_action(self, action: Tensor) -> None:
        self._action = action

    def sample(self) -> Tensor:
        return self.dist.sample().detach()

    def rsample(self) -> Tensor:
        """Sampling by reparameterization trick,
        which means the returned tensor is backwardable
        """
        return self.dist.rsample()

    def eval_action(self, deterministic: bool = False) -> Array:
        """Sample actions for evaluation, which leave no action cache
        and returns numpy array.
        """
        if deterministic:
            act = self.best_action()
        else:
            act = self.sample()
        return act.squeeze_().cpu().numpy()

    @abstractmethod
    def __getitem__(self, idx: Index) -> Self:
        pass

    @abstractmethod
    def best_action(self) -> Tensor:
        pass

    @abstractmethod
    def log_prob(self, use_baction: bool = False) -> Tensor:
        pass

    @abstractmethod
    def entropy(self) -> Tensor:
        pass


class BernoulliPolicy(Policy):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(Bernoulli(*args, **kwargs))
        self.dist: Bernoulli

    def best_action(self) -> Tensor:
        return self.dist.probs > 0.5

    def log_prob(self, use_baction: bool = False) -> Tensor:
        if use_baction:
            return self.dist.log_prob(self.baction())
        else:
            return self.dist.log_prob(self.action())

    def entropy(self) -> Tensor:
        return self.dist.entropy()

    def __getitem__(self, idx: Index) -> Self:
        return BernoulliPolicy(logits=self.dist.logits[idx])


class CategoricalPolicy(Policy):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(Categorical(*args, **kwargs))
        self.dist: Categorical

    def best_action(self) -> Tensor:
        return self.dist.probs.argmax(dim=-1)

    def log_prob(self, use_baction: bool = False) -> Tensor:
        if use_baction:
            return self.dist.log_prob(self.baction())
        else:
            return self.dist.log_prob(self.action())

    def entropy(self) -> Tensor:
        return self.dist.entropy()

    def __getitem__(self, idx: Index) -> Self:
        return CategoricalPolicy(logits=self.dist.logits[idx])


class GaussianPolicy(Policy):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(Normal(*args, **kwargs))
        self.dist: Normal

    def best_action(self) -> Tensor:
        return self.dist.mean

    def entropy(self) -> Tensor:
        return self.dist.entropy().sum(-1)

    def log_prob(self, use_baction: bool = False) -> Tensor:
        if use_baction:
            return self.dist.log_prob(self.baction()).sum(-1)
        else:
            return self.dist.log_prob(self.action()).sum(-1)

    def __getitem__(self, idx: Index) -> Self:
        return GaussianPolicy(self.dist.mean[idx], self.dist.stddev[idx])


class TanhGaussianPolicy(GaussianPolicy):
    def __init__(self, *args, epsilon: float = 1e-6, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._pre_tanh: Optional[Tensor] = None
        self.epsilon = epsilon

    def sample(self) -> Tensor:
        pre_tanh = self.dist.sample().detach()
        self._pre_tanh = pre_tanh
        return torch.tanh(pre_tanh)

    def rsample(self) -> Tensor:
        res = self.dist.rsample()
        self._pre_tanh = res
        return torch.tanh(res)

    def best_action(self) -> Tensor:
        return torch.tanh(self.dist.mean)

    def log_prob(self, use_baction: bool = False) -> Tensor:
        action = self._baction if use_baction else self._action
        if self._pre_tanh is None:
            pre_tanh = torch.log((1.0 + action) / (1.0 - action)) / 2
        else:
            pre_tanh = self._pre_tanh
        log_n = self.dist.log_prob(pre_tanh).sum(-1)
        log_da_du = torch.log(1.0 - action ** 2 + self.epsilon).sum(-1)
        return log_n - log_da_du


class PolicyDist(ABC, nn.Module):
    def __init__(self, action_dim: int, *args, **kwargs) -> None:
        super().__init__()
        self.action_dim = action_dim

    @property
    def input_dim(self) -> int:
        return self.action_dim

    @abstractmethod
    def forward(self, t: Tensor) -> Policy:
        pass


class BernoulliDist(PolicyDist):
    """Bernoulli policy with no learnable parameter
    """

    def forward(self, x: Tensor) -> Policy:
        return BernoulliPolicy(logits=x)


class CategoricalDist(PolicyDist):
    """Categorical policy with no learnable parameter
    """

    def forward(self, x: Tensor) -> Policy:
        return CategoricalPolicy(logits=x)


class GaussinanDist(PolicyDist):
    """Gaussian policy which takes both mean and stdev as inputs
    """

    @property
    def input_dim(self) -> int:
        return self.action_dim * 2

    def forward(self, x: Tensor) -> Policy:
        size = x.size(1) // 2
        mean, stddev = x[:, :size], x[:, size:]
        return GaussianPolicy(mean, F.softplus(stddev))


class TanhGaussianDist(GaussinanDist):
    """Tanh clipped Gaussian policy
    """

    def __init__(
        self,
        action_dim: int,
        logstd_range: Tuple[float, float] = (-20.0, 2.0),
        **kwargs,
    ) -> None:
        super().__init__(action_dim)
        self.logstd_range = logstd_range

    def forward(self, x: Tensor) -> Policy:
        size = x.size(1) // 2
        mu, logstd = x[:, :size], x[:, size:]
        std = torch.exp(logstd.clamp_(*self.logstd_range))
        return TanhGaussianPolicy(mu, std)


class SeparateStdGaussianDist(PolicyDist):
    """Gaussian policy which takes only mean as an input, and has a standard deviation
       independent with states, as a lernable parameter.
    """

    def __init__(self, action_dim: int, device: Device) -> None:
        super().__init__(action_dim)
        self.stddev = nn.Parameter(device.zeros(action_dim))

    def forward(self, x: Tensor) -> Policy:
        return GaussianPolicy(x, F.softplus(self.stddev))



from abc import ABC, abstractmethod
import torch
from torch import LongTensor, Tensor
from torch.optim import Optimizer
from typing import Optional
from ..net.value import DiscreteQFunction
from ..prelude import Array


class Cooler(ABC):
    @abstractmethod
    def __call__(self) -> float:
        pass

    def lr_decay(self, optim: Optimizer) -> None:
        for param_group in optim.param_groups:
            if "lr" in param_group:
                param_group["lr"] = self.__call__()


class LinearCooler(Cooler):
    """decrease epsilon linearly, from initial to minimal, via `max_step` steps
    """

    def __init__(self, initial: float, minimal: float, max_step: int) -> None:
        if initial < minimal:
            raise ValueError(
                f"[LinearCooler.__init__] the minimal value({minimal})"
                " is bigger than the initial value {initial}"
            )
        self.base = initial - minimal
        self.minimal = minimal
        self.max_step = max_step
        self.left = max_step

    def __call__(self) -> float:
        self.left = max(0, self.left - 1)
        return (float(self.left) / self.max_step) * self.base + self.minimal


class DummyCooler(Cooler):
    """Do nothing
    """

    def __init__(self, initial: float, *args) -> None:
        self.initial = initial

    def __call__(self) -> float:
        return self.initial


class Explorer(ABC):
    def select_action(self, state: Array, qfunc: DiscreteQFunction) -> LongTensor:
        return self.select_from_value(qfunc.q_value(state).detach())

    @abstractmethod
    def select_from_value(self, value: Tensor) -> LongTensor:
        pass

    @abstractmethod
    def add_noise(self, action: Tensor) -> Tensor:
        pass


class Greedy(Explorer):
    """deterministic greedy policy
    """

    def select_from_value(self, value: Tensor) -> LongTensor:
        return value.argmax(-1)  # type: ignore

    def add_noise(self, action: Tensor) -> Tensor:
        return action


class EpsGreedy(Explorer):
    """ε-greedy policy
    """

    def __init__(self, epsilon: float, cooler: Cooler) -> None:
        self.epsilon = epsilon
        self.cooler = cooler

    def select_from_value(self, value: Tensor) -> LongTensor:
        old_eps = self.epsilon
        self.epsilon = self.cooler()
        action_dim = value.size(-1)
        greedy = value.argmax(-1).cpu()
        random = torch.randint_like(greedy, 0, action_dim)
        random_pos = torch.empty(greedy.shape).uniform_() < old_eps
        return torch.where(random_pos, random, greedy)

    def add_noise(self, action: Tensor) -> Tensor:
        raise NotImplementedError("We can't use EpsGreedy with continuous action")


class GaussianNoise(Explorer):
    def __init__(
        self, std: Cooler = DummyCooler(0.1), clip: Optional[float] = None
    ) -> None:
        self.std = std
        self.clip = clip

    def add_noise(self, action: Tensor) -> Tensor:
        noise = torch.randn_like(action).mul_(self.std())
        if self.clip is not None:
            noise.clamp_(-self.clip, self.clip)
        return noise + action

    def select_from_value(self, value: Tensor) -> LongTensor:
        raise NotImplementedError()



"""TD3(Twin Delayed Deep Deterministic Policy Gradient)
Paper: https://arxiv.org/abs/1802.09477
"""
from copy import deepcopy
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F
from .base import OneStepAgent
from .ddpg import DDPGAgent
from ..config import Config
from ..prelude import Array
from ..utils.misc import clamp_actions_


class TD3Agent(DDPGAgent):
    def __init__(self, config: Config) -> None:
        OneStepAgent.__init__(self, config)
        self.net = config.net("td3")
        self.target_net = deepcopy(self.net)
        self.actor_opt = config.optimizer(self.net.actor_params(), key="actor")
        self.critic_opt = config.optimizer(self.net.critic_params(), key="critic")
        self.explorer = config.explorer()
        self.target_explorer = config.explorer(key="target")
        self.eval_explorer = config.explorer(key="eval")
        self.replay = config.replay_buffer()
        self.batch_indices = config.device.indices(config.replay_batch_size)
        self.action_range = tuple(torch.from_numpy(t) for t in self.env.spec._act_range)

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        actions = self.target_net.action(next_states)
        actions = self.target_explorer.add_noise(actions)
        clamp_actions_(actions, *self.action_range)
        q1, q2 = self.target_net.q_values(next_states, actions)
        return torch.min(q1, q2)

    def _train(self) -> None:
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_array(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done = map(np.asarray, zip(*obs))
        mask = self.config.device.tensor(1.0 - done)
        q_next = self._q_next(next_states).squeeze_()
        q_target = q_next.mul_(mask * self.config.discount_factor).add_(
            self.config.device.tensor(rewards)
        )
        q1, q2 = self.net.q_values(states, actions)

        #  Backward critic loss
        critic_loss = F.mse_loss(q1.squeeze_(), q_target) + F.mse_loss(
            q2.squeeze_(), q_target
        )
        self._backward(critic_loss, self.critic_opt, self.net.critic_params())

        #  Delayed policy update
        if (self.update_steps + 1) % self.config.policy_update_freq != 0:
            return
        action = self.net.action(states)
        policy_loss = -self.net.q_value(states, action).mean()
        self._backward(policy_loss, self.actor_opt, self.net.actor_params())

        #  Update target network
        self.target_net.soft_update(self.net, self.config.soft_update_coef)



import click
from pathlib import Path
from typing import List, Optional
import warnings
from .agents import Agent, EpisodeResult
from .lib.mpi import IS_MPI_ROOT


class Experiment:
    SAVE_FILE_DEFAULT = "rainy-agent.pth"
    ACTION_FILE_DEFAULT = "actions.json"

    def __init__(
        self,
        ag: Agent,
        save_file_name: Optional[str] = None,
        action_file_name: Optional[str] = None,
    ) -> None:
        self.ag = ag
        self.logger = ag.logger
        self.config = ag.config
        self.logger.summary_setting(
            "train",
            ["episodes", "total_steps", "update_steps"],
            interval=ag.config.episode_log_freq,
            color="red",
        )
        self.logger.summary_setting(
            "eval",
            ["total_steps", "update_steps"],
            interval=ag.config.eval_times,
            color="green",
        )
        action_file = action_file_name or self.ACTION_FILE_DEFAULT
        self._action_file = Path(action_file)
        self._save_file_name = save_file_name or self.SAVE_FILE_DEFAULT
        self.episode_offset = 0

    def log_episode(self, episodes: int, results: List[EpisodeResult]) -> None:
        for i, res in enumerate(results):
            self.logger.submit(
                "train",
                episodes=episodes + i + self.episode_offset,
                total_steps=self.ag.total_steps,
                update_steps=self.ag.update_steps,
                reward=res.reward,
                length=res.length,
            )

    def action_file(self) -> Path:
        return self.logger.logdir.joinpath(
            "{}-{}{}".format(self.action_file.stem, episodes, self.action_file.suffix)
        )

    def log_eval(self, episodes: int, eval_render: bool = False) -> None:
        logdir = self.logger.logdir
        if self.config.save_eval_actions and logdir is not None:
            fname = logdir.joinpath(
                "{}-{}{}".format(action_file.stem, episodes, action_file.suffix)
            )
            results = self._eval_impl(render=eval_render, action_file=fname)
        else:
            results = self._eval_impl(render=eval_render)
        for res in results:
            self.logger.submit(
                "eval",
                total_steps=self.ag.total_steps,
                update_steps=self.ag.update_steps,
                rewards=res.reward,
                length=res.length,
            )

    @staticmethod
    def _has_period(turn: int, width: int, freq: Optional[int]) -> bool:
        return freq and turn != 0 and turn // freq != (turn - width) // freq

    def _save(self, suffix: str = "") -> None:
        self.ag.save(self._save_file_name + suffix, self.logger.logdir)

    def train(self, saveid_start: int = 0, eval_render: bool = False) -> None:
        if not self.logger.ready:
            self.logger.setup()
        logdir = self.logger.logdir.as_posix()
        click.secho(f"Train stared :) Logdir: {logdir}", bg="white", fg="black")

        episodes = 0
        steps = 0
        save_id = saveid_start

        for res in self.ag.train_episodes(self.config.max_steps):
            ep_len = len(res)
            if ep_len > 0 and IS_MPI_ROOT:
                self.log_episode(episodes, res)
            episodes += ep_len
            step_diff = self.ag.total_steps - steps
            steps = self.ag.total_steps

            if not IS_MPI_ROOT:
                continue

            if self._has_period(steps, step_diff, self.config.eval_freq):
                self.log_eval(episodes, eval_render)
            if self._has_period(steps, step_diff, self.config.save_freq):
                self._save(suffix=".{}".format(save_id))
                save_id += 1

        self.log_eval(episodes, eval_render)
        self._save()
        self.ag.close()

    def _eval_impl(
        self,
        render: bool = False,
        replay: bool = False,
        action_file: Optional[str] = None,
    ) -> None:
        n = self.config.eval_times
        self.ag.set_mode(train=False)
        if action_file is not None:
            res = [self.ag.eval_and_save(action_file, render=render) for _ in range(n)]
        elif hasattr(self.ag, "eval_parallel") and not render and not replay:
            res = self.ag.eval_parallel(n)  # type: ignore
        else:
            res = [self.ag.eval_episode(render=render) for _ in range(n)]
        self.ag.set_mode(train=True)
        return res

    def _load_agent(self, logdir: Path) -> bool:
        if p.exists():
            self.ag.load(self._save_file_name, logdir)
            return True
        return False

    def retrain(
        self, logdir_: str, additional_steps: int = 100, eval_render: bool = False
    ) -> None:
        logdir = Path(logdir_)
        if not self._load_agent(logdir):
            raise ValueError("File'{}' does not exists".format(self._save_file_name))
        total_steps, episodes = self.logger.retrive(logdir)
        save_files = list(logdir.glob(self.SAVE_FILE_DEFAULT + ".*"))
        if len(save_files) > 0:
            save_id = len(save_files)
        else:
            save_id = 0
        self.ag.total_steps = total_steps
        self.ag.config.max_steps += additional_steps
        self.episode_offset = episodes
        self.train(save_id, eval_render)

    def evaluate(
        self, logdir_: str, render: bool = False, replay: bool = False
    ) -> None:
        logdir = Path(logdir_)
        if not self._load_agent(logdir):
            raise ValueError("File'{}' does not exists".format(self._save_file_name))
        res = self._eval_impl(render, replay)
        print("{}".format(res))
        if render:
            input("--Press Enter to exit--")
        if replay:
            try:
                self.config.eval_env.unwrapped.replay()
            except Exception:
                warnings.warn("This environment does not support replay")
        self.ag.close()

    def random(
        self,
        render: bool = False,
        replay: bool = False,
        action_file: Optional[str] = None,
    ) -> None:

        if self.config.save_eval_actions:
            action_file = "random-" + self._action_file.as_posix
            res = self.ag.random_and_save(action_file, render=render)
        else:
            res = self.ag.random_episode(render=render)
        print("{}".format(res))
        if replay:
            try:
                self.config.eval_env.unwrapped.replay()
            except Exception:
                warnings.warn("This environment does not support replay")
        self.ag.close()



"""Defines some reusable NN layers, called 'Block'
"""
from abc import ABC
import numpy as np
from torch import nn, Tensor
from typing import List, Sequence, Tuple
from .init import Initializer, orthogonal


class NetworkBlock(nn.Module, ABC):
    """Defines a NN block which returns 1-dimension Tensor
    """

    input_dim: Sequence[int]
    output_dim: int


class DummyBlock(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return x


class LinearHead(NetworkBlock):
    """One FC layer
    """

    def __init__(
        self, input_dim: int, output_dim: int, init: Initializer = Initializer()
    ) -> None:
        super().__init__()
        self.fc: nn.Linear = init(nn.Linear(input_dim, output_dim))  # type: ignore
        self.input_dim = (input_dim,)
        self.output_dim = output_dim

    def forward(self, x: Tensor) -> Tensor:
        return self.fc(x)


class ConvBody(NetworkBlock):
    """Multiple CNN layers + FC
    """

    def __init__(
        self,
        cnns: Sequence[nn.Conv2d],
        fc: nn.Linear,
        input_dim: Tuple[int, int, int],
        activator: nn.Module = nn.ReLU(inplace=True),
        init: Initializer = Initializer(orthogonal(nonlinearity="relu")),
    ) -> None:
        super().__init__()
        self.cnns = init.make_list(cnns)
        self.fc: nn.Linear = init(fc)  # type: ignore
        self.input_dim = input_dim
        self.output_dim = self.fc.out_features
        self.activator = activator

    def forward(self, x: Tensor) -> Tensor:
        for cnn in self.cnns:
            x = self.activator(cnn(x))
        x = x.view(x.size(0), -1)
        x = self.activator(self.fc(x))
        return x


class DQNConv(ConvBody):
    """Convolutuion Network used in https://www.nature.com/articles/nature14236,
       but is parameterized for other usages.
    """

    def __init__(
        self,
        input_dim: Tuple[int, int, int],
        kernel_and_strides: Sequence[Tuple[int, int]] = [(8, 4), (4, 2), (3, 1)],
        hidden_channels: Sequence[int] = (32, 64, 64),
        output_dim: int = 512,
        activator: nn.Module = nn.ReLU(inplace=True),
        init: Initializer = Initializer(orthogonal(nonlinearity="relu")),
    ) -> None:
        in_channel, width, height = input_dim
        cnns, hidden = make_cnns(input_dim, kernel_and_strides, hidden_channels)
        fc = nn.Linear(hidden, output_dim)
        self._output_dim = output_dim
        super().__init__(cnns, fc, input_dim, activator=activator, init=init)


class ResBlock(nn.Sequential):
    def __init__(
        self, channel: int, stride: int = 1, use_batch_norm: bool = True,
    ) -> None:
        super().__init__(
            nn.ReLU(inplace=True),
            self._conv3x3(channel, channel, stride),
            self._batch_norm(use_batch_norm, channel),
            nn.ReLU(inplace=True),
            self._conv3x3(channel, channel, stride),
            self._batch_norm(use_batch_norm, channel),
        )

    @staticmethod
    def _batch_norm(use_batch_norm: bool, channel: int) -> nn.Module:
        if use_batch_norm:
            return nn.BatchNorm2d(channel)
        else:
            return DummyBlock()

    @staticmethod
    def _conv3x3(in_channel: int, out_channel: int, stride: int = 1) -> nn.Conv2d:
        return nn.Conv2d(
            in_channel,
            out_channel,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )

    def forward(self, x: Tensor) -> Tensor:
        residual = x
        out = super().forward(x)
        return out + residual


class ResNetBody(NetworkBlock):
    """Convolutuion Network used in IMPALA
    """

    def __init__(
        self,
        input_dim: Tuple[int, int, int],
        channels: List[int] = [16, 32, 32],
        maxpools: List[tuple] = [(3, 2, 1)] * 3,
        use_batch_norm: bool = True,
        fc_out: int = 256,
        init: Initializer = Initializer(orthogonal(nonlinearity="relu")),
    ) -> None:
        def layer(in_channel: int, out_channel: int, maxpool: tuple) -> nn.Sequential:
            return nn.Sequential(
                ResBlock._conv3x3(in_channel, out_channel),
                ResBlock._batch_norm(use_batch_norm, out_channel),
                nn.MaxPool2d(*maxpool),
                ResBlock(out_channel, use_batch_norm=use_batch_norm),
                ResBlock(out_channel, use_batch_norm=use_batch_norm),
            )

        super().__init__()

        self.input_dim = input_dim
        self.output_dim = fc_out
        self.res_blocks = init.make_list(
            [layer(*t) for t in zip([input_dim[0]] + channels, channels, maxpools)]
        )
        self.relu = nn.ReLU(inplace=True)
        conved = calc_cnn_hidden(maxpools, *input_dim[1:])
        fc_in = np.prod((channels[-1], *conved))
        self.fc = nn.Linear(fc_in, fc_out)

    def forward(self, x: Tensor) -> Tensor:
        for block in self.res_blocks:
            x = block(x)
        x = self.relu(x)
        x = self.fc(x.view(x.size(0), -1))
        return self.relu(x)


class FcBody(NetworkBlock):
    def __init__(
        self,
        input_dim: int,
        units: List[int] = [64, 64],
        activator: nn.Module = nn.ReLU(inplace=True),
        init: Initializer = Initializer(),
    ) -> None:
        super().__init__()
        self.input_dim = (input_dim,)
        self.output_dim = units[-1]
        dims = [input_dim] + units
        self.layers = init.make_list(
            map(lambda i, o: nn.Linear(i, o), *zip(dims[:-1], dims[1:]))
        )
        self.activator = activator

    def forward(self, x: Tensor) -> Tensor:
        for layer in self.layers:
            x = self.activator(layer(x))
        return x


def make_cnns(
    input_dim: Tuple[int, int, int],
    params: Sequence[tuple],
    hidden_channels: Sequence[int],
) -> Tuple[List[nn.Conv2d], int]:
    """Make a list of CNNs from lists of parameters.
    """
    channel, width, height = input_dim
    hiddens = list(hidden_channels)
    res = []
    for ic, oc, param in zip([channel] + hiddens, hiddens, params):
        res.append(nn.Conv2d(ic, oc, *param))
    hidden = (hidden_channels[-1], *calc_cnn_hidden(params, height, width))
    return res, np.prod(hidden)


def calc_cnn_hidden(
    params: Sequence[tuple], height: int, width: int
) -> Tuple[int, int]:
    """Calcurate hidden dim of a CNN.
       See https://pytorch.org/docs/stable/nn.html#torch.nn.Conv2d for detail.
    """
    for param in params:
        kernel, stride, pad = param if len(param) > 2 else (*param, 0)
        if isinstance(pad, int):
            h_pad, w_pad = pad, pad
        else:
            h_pad, w_pad = pad
        height = (height - kernel + 2 * h_pad) // stride + 1
        width = (width - kernel + 2 * w_pad) // stride + 1
    assert width > 0 and height > 0, "Convolution makes dim < 0!!!"
    return height, width



from functools import partial
import numpy as np
from typing import Callable, Iterable, Optional
import torch
from torch import nn, Tensor

InitFn = Callable[[Tensor], Tensor]


def uniform(min: float = -1.0, max: float = 1.0) -> InitFn:
    return partial(nn.init.uniform_, a=min, b=max)


def normal(mean: float = 0.0, std: float = 1.0) -> InitFn:
    return partial(nn.init.normal_, mean=mean, std=std)


def orthogonal(gain: float = 1.0, nonlinearity: Optional[str] = None) -> InitFn:
    if nonlinearity is not None:
        gain = nn.init.calculate_gain(nonlinearity)
    return partial(nn.init.orthogonal_, gain=gain)


def kaiming_normal(**kwargs) -> InitFn:
    return partial(nn.init.kaiming_normal_, **kwargs)


def kaiming_uniform(**kwargs) -> InitFn:
    return partial(nn.init.kaiming_uniform_, **kwargs)


def xavier_uniform(**kwargs) -> InitFn:
    return partial(nn.init.xavier_uniform_, **kwargs)


def fanin_uniform() -> InitFn:
    def _fanin_uniform(w: Tensor) -> Tensor:
        if w.dim() <= 2:
            fan_in = w.size(0)
        else:
            fan_in = np.prod(w.shape[1:])
        bound = 1.0 / np.sqrt(fan_in)
        return nn.init.uniform_(w, -bound, bound)

    return _fanin_uniform


def constant(val: float) -> InitFn:
    return partial(nn.init.constant_, val=val)


def zero() -> InitFn:
    return partial(nn.init.constant_, val=0)


def lstm_bias(forget: float = 1.0, other: float = 0.0) -> InitFn:
    """Set forget bias and others separately.
    """

    @torch.no_grad()
    def __set_bias(t: Tensor) -> Tensor:
        i = t.size(0) // 4
        t.fill_(other)
        return t[i : 2 * i].fill_(forget)

    return __set_bias


class Initializer:
    """Utility Class to initialize weight parameters of NN
    """

    def __init__(
        self,
        weight_init: InitFn = orthogonal(),
        bias_init: InitFn = zero(),
        scale: float = 1.0,
    ) -> None:
        """If nonlinearity is specified, use orthogonal
           with calucurated gain by torch.init.calculate_gain.
        """
        self.weight_init = weight_init
        self.bias_init = bias_init
        self.scale = scale

    def __call__(self, mod: nn.Module) -> nn.Module:
        return self.__init_dispatch(mod)

    def make_list(self, mods: Iterable[nn.Module]) -> nn.ModuleList:
        return nn.ModuleList([self.__init_dispatch(mod) for mod in mods])

    def __init_dispatch(self, mod: nn.Module) -> nn.Module:
        if isinstance(mod, nn.Sequential) or isinstance(mod, nn.ModuleList):
            for child in mod.children():
                self.__init_dispatch(child)
        else:
            self.__init_mod(mod)
        return mod

    def __init_mod(self, mod: nn.Module) -> nn.Module:
        if isinstance(mod, nn.BatchNorm2d):
            return self.__init_batch_norm(mod)
        for name, param in mod.named_parameters():
            if "weight" in name:
                self.weight_init(param)
            elif "bias" in name:
                self.bias_init(param)
        return mod

    def __init_batch_norm(self, mod: nn.BatchNorm2d) -> nn.Module:
        mod.weight.data.fill_(1.0)
        mod.bias.data.zero_()
        return mod



"""
This module has an implementation of Bootstrapped DQN, which is described in
- Deep Exploration via Bootstrapped DQN(https://arxiv.org/abs/1602.04621)
- Randomized Prior Functions for Deep Reinforcement Learning(https://arxiv.org/abs/1806.03335)
"""
import copy
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F
from typing import Tuple
from .base import OneStepAgent
from ..config import Config
from ..prelude import Action, Array, State
from ..replay import BootDQNReplayFeed


class EpisodicBootDQNAgent(OneStepAgent):
    SAVED_MEMBERS = "net", "policy", "total_steps"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        if not self.env.spec.is_discrete():
            raise RuntimeError("DQN only supports discrete action space.")
        self.net = config.net("bootdqn")
        self.optimizer = config.optimizer(self.net.parameters())
        self.policy = config.explorer()
        self.eval_policy = config.explorer(key="eval")
        self.replay = config.replay_buffer()
        self.replay.allow_overlap = True
        self.active_head = 0
        if self.replay.feed is not BootDQNReplayFeed:
            raise RuntimeError("BootDQNAgent needs BootDQNReplayFeed")

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    @torch.no_grad()
    def eval_action(self, state: Array) -> Action:
        return self.eval_policy.select_action(state, self.net).item()  # type: ignore

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        with torch.no_grad():
            qs = self.net.q_i_s(self.active_head, self.env.extract(state)).detach()
        action = self.policy.select_from_value(qs).item()
        next_state, reward, done, info = self.env.step(action)
        self._append_to_replay(state, action, reward, next_state, done)
        if done:
            self._train()
            self.active_head = np.random.randint(self.config.num_ensembles)
        return next_state, reward, done, info

    def _append_to_replay(self, *transition) -> None:
        n_ens = self.config.num_ensembles
        mask = np.random.uniform(0, 1, n_ens) < self.config.replay_prob
        self.replay.append(*transition, mask)

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        return self.net(next_states).max(axis=-1)[0]

    def _train(self) -> None:
        gamma = self.config.discount_factor
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_array(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done, mask = map(np.asarray, zip(*obs))
        q_next = self._q_next(next_states)
        r = self.tensor(rewards).view(-1, 1)
        q_target = r + q_next * self.tensor(1.0 - done).mul_(gamma).view(-1, 1)
        q_current = self.net.q_s_a(states, actions)
        loss = F.mse_loss(q_current, q_target, reduction="none")
        masked_loss = loss.masked_select(self.tensor(mask, dtype=torch.bool)).mean()
        self._backward(masked_loss, self.optimizer, self.net.parameters())
        self.network_log(q_value=q_current.mean().item(), value_loss=masked_loss.item())


class BootDQNAgent(EpisodicBootDQNAgent):
    SAVED_MEMBERS = "net", "policy", "total_steps", "target_net"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.target_net = copy.deepcopy(self.net)
        self.replay.allow_overlap = False

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        train_started = self.total_steps > self.config.train_start
        if train_started:
            with torch.no_grad():
                qs = self.net.q_i_s(self.active_head, self.env.extract(state)).detach()
            action = self.policy.select_from_value(qs).item()
        else:
            action = self.env.spec.random_action()
        next_state, reward, done, info = self.env.step(action)
        self._append_to_replay(state, action, reward, next_state, done)
        if done:
            self.active_head = np.random.randint(self.config.num_ensembles)
        if train_started:
            self._train()
        return next_state, reward, done, info

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        return self.net(next_states).max(axis=-1)[0]

    def _train(self) -> None:
        super()._train
        if (self.update_steps + 1) % self.config.sync_freq == 0:
            self.target_net.load_state_dict(self.net.state_dict())



from abc import ABC, abstractmethod
import copy
import numpy as np
from pathlib import Path
import torch
from torch import nn, Tensor
from typing import (
    Callable,
    ClassVar,
    Generic,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)
import warnings
from ..config import Config
from ..lib import mpi
from ..net import DummyRnn, RnnState
from ..envs import EnvExt
from ..prelude import Action, Array, ArrayLike, State


class EpisodeResult(NamedTuple):
    reward: np.float32
    length: np.int32

    def __repr__(self) -> str:
        return "EpisodeResult(reward: {}, length: {})".format(self.reward, self.length)


class Agent(ABC):
    """Children must call super().__init__(config) first
    """

    SAVED_MEMBERS: ClassVar[Sequence[str]]

    def __init__(self, config: Config) -> None:
        self.config = config
        self.logger = config.logger
        self.env = config.env()
        self.total_steps = 0
        self.logger.summary_setting(
            "network",
            ["total_steps", "update_steps"],
            interval=config.network_log_freq,
            color="blue",
        )

    @abstractmethod
    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        """Train the agent.
        """
        pass

    @abstractmethod
    def eval_action(self, state: Array) -> Action:
        """Return the best action according to training results.
        """
        pass

    @property
    @abstractmethod
    def update_steps(self) -> int:
        pass

    def eval_reset(self) -> None:
        pass

    def set_mode(self, train: bool = True) -> None:
        pass

    def network_log(self, **kwargs) -> None:
        kwargs["total_steps"] = self.total_steps
        kwargs["update_steps"] = self.update_steps
        self.logger.submit("network", **kwargs)

    def close(self) -> None:
        self.env.close()
        self.logger.close()

    def __eval_episode(
        self, select_action: Callable[[Array], Action], render: bool
    ) -> Tuple[EpisodeResult, EnvExt]:
        total_reward = 0.0
        steps = 0
        env = self.config.eval_env
        if self.config.seed is not None:
            env.seed(self.config.seed)
        state = env.reset()
        while True:
            if render:
                env.render()
            state = env.extract(state)
            action = select_action(state)
            state, reward, done, info = env.step(action)
            steps += 1
            total_reward += reward
            res = self._result(done, info, total_reward, steps)
            if res is not None:
                self.eval_reset()
                return (res, env)

    def _result(
        self, done: bool, info: dict, total_reward: float, episode_length: int,
    ) -> Optional[EpisodeResult]:
        if self.env.use_reward_monitor:
            if "episode" in info:
                return EpisodeResult(info["episode"]["r"], info["episode"]["l"])
        elif done:
            return EpisodeResult(total_reward, episode_length)
        return None

    def random_episode(self, render: bool = False) -> EpisodeResult:
        def act(_state) -> Action:
            return self.config.eval_env.spec.random_action()

        return self.__eval_episode(act, render)[0]

    def random_and_save(self, fname: str, render: bool = False) -> EpisodeResult:
        def act(_state) -> Action:
            return self.config.eval_env.spec.random_action()

        res, env = self.__eval_episode(act, render)
        env.save_history(fname)
        return res

    def eval_episode(self, render: bool = False) -> EpisodeResult:
        return self.__eval_episode(self.eval_action, render)[0]

    def eval_and_save(self, fname: str, render: bool = False) -> EpisodeResult:
        res, env = self.__eval_episode(self.eval_action, render)
        env.save_history(fname)
        return res

    def save(self, filename: str, directory: Optional[Path] = None) -> None:
        if not mpi.IS_MPI_ROOT:
            return None
        save_dict = {}
        for member_str in self.SAVED_MEMBERS:
            value = getattr(self, member_str)
            if isinstance(value, nn.DataParallel):
                save_dict[member_str] = value.module.state_dict()
            elif hasattr(value, "state_dict"):
                save_dict[member_str] = value.state_dict()
            else:
                save_dict[member_str] = value
        if directory is None:
            directory = self.logger.logdir
        torch.save(save_dict, directory.joinpath(filename))

    def load(self, filename: str, directory: Optional[Path] = None) -> bool:
        if not mpi.IS_MPI_ROOT:
            return False
        if directory is None:
            directory = self.logger.logdir
        path = directory.joinpath(filename)
        if not path.exists():
            return False
        saved_dict = torch.load(path, map_location=self.config.device.unwrapped)
        #  For backward compatibility, we need to check both index and name
        for idx, member_str in enumerate(self.SAVED_MEMBERS):
            if idx in saved_dict:
                saved_item = saved_dict[idx]
            elif member_str in saved_dict:
                saved_item = saved_dict[member_str]
            else:
                warnings.warn("Member {} wasn't loaded".format(member_str))
                continue
            mem = getattr(self, member_str)
            if isinstance(mem, nn.DataParallel):
                mem.module.load_state_dict(saved_item)
            elif hasattr(mem, "state_dict"):
                mem.load_state_dict(saved_item)
            else:
                setattr(self, member_str, saved_item)
        return True

    def tensor(self, arr: ArrayLike, dtype: torch.dtype = torch.float32) -> Tensor:
        """A shorthand method for calling `device.tensor`.
        """
        return self.config.device.tensor(arr, dtype)

    def _backward(
        self,
        loss: Tensor,
        opt: torch.optim.Optimizer,
        params: Optional[Iterable[Tensor]] = None,
    ) -> None:
        opt.zero_grad()
        loss.backward()
        if params is not None and self.config.grad_clip is not None:
            nn.utils.clip_grad_norm_(params, self.config.grad_clip)
        opt.step()


class OneStepAgent(Agent, Generic[State]):
    @abstractmethod
    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        pass

    @property
    def update_steps(self) -> int:
        return max(0, self.total_steps - self.config.train_start)

    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        if self.config.seed is not None:
            self.env.seed(self.config.seed)
        state = self.env.reset()
        total_reward = 0.0
        episode_length = 0
        while True:
            state, reward, done, info = self.step(state)
            self.total_steps += 1
            total_reward += reward
            episode_length += 1
            res = self._result(done, info, total_reward, episode_length)
            if res is not None:
                yield [res]
                state = self.env.reset()
                total_reward = 0.0
                episode_length = 0
            if self.total_steps >= max_steps:
                break


class NStepParallelAgent(Agent, Generic[State]):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.rewards = np.zeros(config.nworkers, dtype=np.float32)
        self.episode_length = np.zeros(config.nworkers, dtype=np.int)
        self.episode_results: List[EpisodeResult] = []
        self.penv = config.parallel_env()
        self.eval_rnns: RnnState = DummyRnn.DUMMY_STATE
        self.step_width = self.config.nsteps * self.config.nworkers * mpi.global_size()

    def eval_parallel(
        self, n: Optional[int] = None, entropy: Optional[Array[float]] = None,
    ) -> List[EpisodeResult]:
        reserved = (
            copy.deepcopy(self.rewards),
            copy.deepcopy(self.episode_length),
            copy.deepcopy(self.episode_results),
        )
        self.rewards.fill(0.0)
        self.episode_length.fill(0)
        self.episode_results.clear()
        if n is None:
            n = self.config.nworkers
        self.config.set_parallel_seeds(self.penv)

        states = self.penv.reset()
        mask = self.config.device.ones(self.config.nworkers)
        while True:
            actions = self.eval_action_parallel(
                self.penv.extract(states), mask, entropy
            )
            states, rewards, done, info = self.penv.step(actions)
            self.episode_length += 1
            self.rewards += rewards
            self.report_reward(done, info)
            if n <= len(self.episode_results):
                break

        res = self.episode_results
        self.rewards, self.episode_length, self.episode_results = reserved
        self.eval_reset()
        return res

    @abstractmethod
    def eval_action_parallel(
        self, states: Array, mask: torch.Tensor, ent: Optional[Array[float]] = None,
    ) -> Array[Action]:
        pass

    @abstractmethod
    def nstep(self, states: Array[State]) -> Array[State]:
        pass

    @abstractmethod
    def _reset(self, initial_states: Array[State]) -> None:
        pass

    @property
    def update_steps(self) -> int:
        return self.total_steps // self.step_width

    def rnn_init(self) -> RnnState:
        return DummyRnn.DUMMY_STATE

    def report_reward(self, done: Array[bool], info: Array[dict]) -> None:
        if self.penv.use_reward_monitor:
            for i in filter(lambda i: "episode" in i, info):
                self.episode_results.append(
                    EpisodeResult(i["episode"]["r"], i["episode"]["l"])
                )
        else:
            for i in filter(lambda i: done[i], range(self.config.nworkers)):
                self.episode_results.append(
                    EpisodeResult(self.rewards[i], self.episode_length[i])
                )
                self.rewards[i] = 0.0
                self.episode_length[i] = 0

    def close(self) -> None:
        self.env.close()
        self.penv.close()
        self.logger.close()

    def train_episodes(self, max_steps: int) -> Iterable[List[EpisodeResult]]:
        self.config.set_parallel_seeds(self.penv)
        states = self.penv.reset()
        self._reset(states)
        while True:
            states = self.nstep(states)
            self.total_steps += self.step_width
            if self.episode_results:
                yield self.episode_results
                self.episode_results = []
            if self.total_steps >= max_steps:
                break



from .a2c import A2CAgent
from .acktr import ACKTRAgent
from .aoc import AOCAgent
from .base import Agent, EpisodeResult, NStepParallelAgent, OneStepAgent
from .bootdqn import BootDQNAgent, EpisodicBootDQNAgent
from .ddpg import DDPGAgent
from .dqn import DQNAgent, DoubleDQNAgent, EpisodicDQNAgent
from .ppo import PPOAgent
from .sac import SACAgent
from .td3 import TD3Agent



from torch import Tensor
from typing import Any, Iterable, List, Sequence, Tuple, TypeVar, Union


try:
    from typing import GenericMeta, NamedTupleMeta  # type: ignore

    class GenericNamedMeta(NamedTupleMeta, GenericMeta):
        pass


except ImportError:
    from typing import NamedTupleMeta  # type: ignore

    GenericNamedMeta = NamedTupleMeta  # type: ignore

T = TypeVar("T")
Self = Any
Index = Union[None, int, slice, Tensor, List[Any], Tuple[Any, ...]]
Params = Union[Iterable[Tensor], dict]


class Array(Sequence[T]):
    @property
    def shape(self) -> tuple:
        ...

    def squeeze(self) -> Self:
        ...

    def transpose(self, *args) -> Self:
        ...

    def mean(self, **kwargs) -> Self:
        ...

    def var(self, **kwargs) -> Self:
        ...

    def __rsub__(self, value: Any) -> Self:
        ...

    def __truediv__(self, rvalue: Any) -> Self:
        ...


ArrayLike = Union[Array[Any], List[Any], Tensor]

Action = TypeVar("Action", int, Array)
State = TypeVar("State")



from copy import deepcopy
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F
from typing import Tuple
from .base import OneStepAgent
from ..config import Config
from ..prelude import Action, Array, State
from ..replay import DQNReplayFeed


class EpisodicDQNAgent(OneStepAgent):
    """A DQN variant, which has no target network and
    updates the target only once per episode.
    """

    SAVED_MEMBERS = "net", "policy", "total_steps"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        if not self.env.spec.is_discrete():
            raise RuntimeError("DQN only supports discrete action space.")
        self.net = config.net("dqn")
        self.optimizer = config.optimizer(self.net.parameters())
        self.policy = config.explorer()
        self.eval_policy = config.explorer(key="eval")
        self.replay = config.replay_buffer()
        if self.replay.feed is not DQNReplayFeed:
            raise RuntimeError("DQNAgent needs DQNReplayFeed")
        self.batch_indices = config.device.indices(config.replay_batch_size)

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    @torch.no_grad()
    def eval_action(self, state: Array) -> Action:
        return self.eval_policy.select_action(state, self.net).item()  # type: ignore

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        action = self.policy.select_action(self.env.extract(state), self.net).item()
        next_state, reward, done, info = self.env.step(action)
        self.replay.append(state, action, reward, next_state, done)
        if done:
            self._train()
        return next_state, reward, done, info

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        return self.net(next_states).max(axis=-1)[0]

    def _train(self) -> None:
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_array(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done = map(np.asarray, zip(*obs))
        q_next = self._q_next(next_states).mul_(self.tensor(1.0 - done))
        q_target = self.tensor(rewards).add_(q_next.mul_(self.config.discount_factor))
        q_current = self.net(states)[self.batch_indices, actions]
        loss = F.mse_loss(q_current, q_target)
        self._backward(loss, self.optimizer, self.net.parameters())
        self.network_log(q_value=q_current.mean().item(), value_loss=loss.item())


class DQNAgent(EpisodicDQNAgent):

    SAVED_MEMBERS = "net", "policy", "total_steps", "target_net"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.target_net = deepcopy(self.net)

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        train_started = self.total_steps > self.config.train_start
        if train_started:
            action = self.policy.select_action(self.env.extract(state), self.net).item()
        else:
            action = self.env.spec.random_action()
        next_state, reward, done, info = self.env.step(action)
        self.replay.append(state, action, reward, next_state, done)
        if train_started:
            self._train()
        return next_state, reward, done, info

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        return self.target_net(next_states).max(axis=-1)[0]

    def _train(self):
        super()._train()
        if (self.update_steps + 1) % self.config.sync_freq == 0:
            self.target_net.load_state_dict(self.net.state_dict())


class DoubleDQNAgent(DQNAgent):
    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        """Returns Q values of next_states, supposing torch.no_grad() is called
        """
        q_next = self.target_net(next_states)
        q_value = self.net.q_value(next_states, nostack=True)
        return q_next[self.batch_indices, q_value.argmax(dim=-1)]



import io
import os
import re
from setuptools import find_packages, setup


NAME = "rainy"
AUTHOR = "Yuji Kanagawa"
EMAIL = "yuji.kngw.80s.revive@gmail.com"
URL = "https://github.com/kngwyu/Rainy"
REQUIRES_PYTHON = ">=3.6.0"
DESCRIPTION = "Algorithm and utilities for deep reinforcement learning"

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, "rainy/__init__.py"), "rt", encoding="utf8") as f:
    VERSION = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        LONG_DESCRIPTION = "\n" + f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = DESCRIPTION


REQUIRED = [
    "click>=7.0",
    "colorama>=0.4",
    "GitPython>=2.0",
    "gym[atari]>=0.11.0",
    "numpy>=1.15.0",
    "opencv-python>=3.4",
    "pandas>=0.25",
    "Pillow>=6.2",
    "torch>=1.0",
]
EXTRA = {
    "bullet": ["pybullet>=2.4"],
    "horovod": ["horovod>=0.16"],
}


setup(
    name=NAME,
    version=VERSION,
    url=URL,
    project_urls={"Code": URL, "Issue tracker": URL + "/issues",},
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRED,
    extras_require=EXTRA,
    license="Apache2",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)



"""SAC(Soft Actor Critic) agent
Paper: https://arxiv.org/abs/1812.05905
"""
from abc import ABC, abstractmethod
import numpy as np
from torch import nn, Tensor
import torch
from torch.nn import functional as F
from typing import Sequence, Tuple
from .base import OneStepAgent
from ..config import Config
from ..net import Policy, SeparatedSACNet
from ..prelude import Action, Array, State


class EntropyTuner(ABC):
    @abstractmethod
    def alpha(self, _: Tensor) -> float:
        pass


class DummyEntropyTuner(EntropyTuner):
    def __init__(self, alpha: float) -> None:
        self._alpha = alpha

    def alpha(self, _: Tensor) -> float:
        return self._alpha


class TrainableEntropyTuner(EntropyTuner):
    def __init__(self, target_entropy: float, config: Config) -> None:
        self.log_alpha = nn.Parameter(config.device.zeros(1))
        self.optim = config.optimizer([self.log_alpha], key="entropy")
        self.target_entropy = target_entropy

    def alpha(self, log_pi: Tensor) -> float:
        alpha_loss = -(self.log_alpha * (log_pi.detach() + self.target_entropy)).mean()
        self.optim.zero_grad()
        alpha_loss.backward()
        self.optim.step()
        return self.log_alpha.detach().exp().item()


class SACAgent(OneStepAgent):
    SAVED_MEMBERS = "net", "target_net", "actor_opt", "critic_opt", "replay"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.net: SeparatedSACNet = config.net("sac")  # type: ignore
        self.target_net = self.net.get_target()
        self.actor_opt = config.optimizer(self.net.actor_params(), key="actor")
        self.critic_opt = config.optimizer(self.net.critic_params(), key="critic")
        self.replay = config.replay_buffer()
        self.batch_indices = config.device.indices(config.replay_batch_size)

        if self.config.automatic_entropy_tuning:
            target = self._target_entropy()
            self.entropy_tuner: EntropyTuner = TrainableEntropyTuner(target, config)
        else:
            self.entropy_tuner = DummyEntropyTuner(self.config.fixed_alpha)

    def _target_entropy(self):
        if self.config.target_entropy is None:
            return -np.prod(self.env.state_dim).item() - 0.0
        else:
            return self.config.target_entropy

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    @torch.no_grad()
    def eval_action(self, state: Array) -> Action:
        policy = self.net.policy(state)
        return policy.eval_action(self.config.eval_deterministic)

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        train_started = self.total_steps > self.config.train_start
        if train_started:
            with torch.no_grad():
                policy = self.net.policy(state)
                action = policy.action().cpu().numpy()
        else:
            action = self.env.spec.random_action()
        action = self.env.spec.clip_action(action)
        next_state, reward, done, info = self.env.step(action)
        self.replay.append(
            state, action, reward * self.config.reward_scale, next_state, done
        )
        if train_started:
            self._train()
        return next_state, reward, done, info

    def _logpi_and_q(self, states: Tensor, policy: Policy) -> Tuple[Tensor, Tensor]:
        actions = policy.baction()
        q1, q2 = self.net.q_values(states, actions)
        return policy.log_prob(use_baction=True), torch.min(q1, q2)

    @torch.no_grad()
    def _q_next(self, next_states: Tensor, alpha: float) -> Tensor:
        policy = self.net.policy(next_states)
        q1, q2 = self.target_net.q_values(next_states, policy.action())
        return torch.min(q1, q2).squeeze_() - alpha * policy.log_prob()

    def _train(self) -> None:
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_array(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done = map(np.asarray, zip(*obs))
        q1, q2, policy = self.net(states, actions)

        # Backward policy loss
        log_pi, new_q = self._logpi_and_q(states, policy)
        alpha = self.entropy_tuner.alpha(log_pi)
        policy_loss = (alpha * log_pi - new_q).mean()
        self._backward(policy_loss, self.actor_opt, self.net.actor_params())

        #  Backward critic loss
        mask = self.config.device.tensor(1.0 - done)
        q_next = self._q_next(next_states, alpha)
        q_target = q_next.mul_(mask * self.config.discount_factor).add_(
            self.config.device.tensor(rewards)
        )
        critic_loss = F.mse_loss(q1.squeeze_(), q_target) + F.mse_loss(
            q2.squeeze_(), q_target
        )
        self._backward(critic_loss, self.critic_opt, self.net.critic_params())

        #  Update target network
        if (self.update_steps + 1) % self.config.sync_freq == 0:
            self.target_net.soft_update(self.net, self.config.soft_update_coef)



"""Dummy environment and utitlities for testing
"""
from enum import Enum
from gym.spaces import Discrete
import numpy as np
from typing import Sequence, Tuple
from .ext import EnvExt, EnvSpec
from ..prelude import Array

ACTION_DIM = 10


class State(Enum):
    START = 0
    ROAD1 = 1
    ROAD2 = 2
    DEATH = 3
    GOAL = 4

    def is_end(self) -> bool:
        return self == State.DEATH or self == State.GOAL

    def to_array(self, dim: tuple) -> np.array:
        return np.repeat(float(self.value) + 1.0, np.prod(dim)).reshape(dim)


class DummyEnv(EnvExt):
    def __init__(
        self, array_dim: Sequence[int] = (16, 16), flatten: bool = False
    ) -> None:
        self.state = State.START
        self.transition = [
            [0.0, 0.7, 0.3, 0.0, 0.0],
            [0.0, 0.0, 0.8, 0.2, 0.0],
            [0.0, 0.0, 0.0, 0.4, 0.6],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
        ]
        self.rewards = [0.0, 0.0, 0.0, -10.0, 20.0]
        self.array_dim = array_dim
        self.flatten = flatten
        if self.flatten:
            dim: Sequence[int] = (np.prod(self.array_dim),)
        else:
            dim = self.array_dim
        self.spec = EnvSpec(dim, Discrete(ACTION_DIM))

    def reset(self) -> State:
        self.state = State.START
        return self.state

    def step(self, _action) -> Tuple[State, float, bool, dict]:
        prob = np.asarray(self.transition[self.state.value])
        self.state = State(np.random.choice(np.arange(5), 1, p=prob))
        return self.state, self.rewards[self.state.value], self.state.is_end(), {}

    def seed(self, int) -> None:
        pass

    def close(self) -> None:
        pass

    def extract(self, state: State) -> Array:
        res = state.to_array(self.array_dim)
        if self.flatten:
            return res.flatten()
        else:
            return res


class DummyEnvDeterministic(DummyEnv):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.transition = [
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
        ]



from gym import spaces, logger
from gym.envs.classic_control import CartPoleEnv
import numpy as np

F32_MAX = np.finfo(np.float32).max


class CartPoleSwingUp(CartPoleEnv):
    START_POSITIONS = ["arbitary", "bottom"]
    ACT_TO_FORCE = [-1.0, 1.0, 0.0]

    def __init__(
        self,
        start_position="arbitary",
        height_threshold=0.5,
        theta_dot_threshold=1.0,
        x_reward_threshold=1.0,
        # This is 2.4 in the original CartPole
        x_threshold=3.0,
        # Aloow 'No operation for action'
        allow_noop=False,
        move_cost=0.1,
    ):
        super().__init__()
        self.x_threshold = x_threshold
        self.start_position = self.START_POSITIONS.index(start_position)
        self._height_threshold = height_threshold
        self._theta_dot_threshold = theta_dot_threshold
        self._x_reward_threshold = x_reward_threshold
        self._move_cost = move_cost
        if allow_noop:
            self.action_space = spaces.Discrete(3)
        self.allow_noop = allow_noop
        high = np.array([1.0, F32_MAX, 1.0, 1.0, F32_MAX])
        self.observation_space = spaces.Box(-high, high, dtype=np.float32)

    def step(self, action):
        """
        action: int
        """
        if not self.action_space.contains(action):
            raise ValueError(f"Invalid action: {action}")
        force = self.force_mag * self.ACT_TO_FORCE[action]
        if self.allow_noop and action != 2:
            move_cost = self._move_cost
        else:
            move_cost = 0.0
        state = self.state
        x, x_dot, theta, theta_dot = state
        costheta, sintheta = np.cos(theta), np.sin(theta)
        temp = (
            force + self.polemass_length * theta_dot * theta_dot * sintheta
        ) / self.total_mass
        thetaacc = (self.gravity * sintheta - costheta * temp) / (
            self.length
            * (4.0 / 3.0 - self.masspole * costheta * costheta / self.total_mass)
        )
        xacc = temp - self.polemass_length * thetaacc * costheta / self.total_mass
        if self.kinematics_integrator == "euler":
            x = x + self.tau * x_dot
            x_dot = x_dot + self.tau * xacc
            theta = theta + self.tau * theta_dot
            theta_dot = theta_dot + self.tau * thetaacc
        else:  # semi-implicit euler
            x_dot = x_dot + self.tau * xacc
            x = x + self.tau * x_dot
            theta_dot = theta_dot + self.tau * thetaacc
            theta = theta + self.tau * theta_dot

        self.state = x, x_dot, theta, theta_dot
        done = bool(x < -self.x_threshold or x > self.x_threshold)

        def _reward():
            is_upright = np.cos(theta) > self._height_threshold
            is_upright &= np.abs(theta_dot) < self._theta_dot_threshold
            is_upright &= np.abs(x) < self._x_reward_threshold
            return 1.0 if is_upright else 0.0 - move_cost

        if not done:
            reward = _reward()
        elif self.steps_beyond_done is None:
            # Pole just fell!
            self.steps_beyond_done = 0
            reward = _reward()
        else:
            if self.steps_beyond_done == 0:
                logger.warn("You are calling 'step()' after the episode ending.")
            self.steps_beyond_done += 1
            reward = 0.0

        return self._obs(), reward, done, {}

    def reset(self):
        self.state = self.np_random.uniform(-0.05, 0.05, size=(4,))
        if self.start_position == 0:
            self.state[2] = self.np_random.uniform(-np.pi, np.pi)
        else:
            self.state[2] += np.pi
        self.steps_beyond_done = None
        return self._obs()

    def _obs(self):
        x, x_dot, theta, theta_dot = self.state
        obs = np.zeros(5, dtype=np.float32)
        obs[0] = x / self.x_threshold
        obs[1] = x_dot / self.x_threshold
        obs[2] = np.sin(theta)
        obs[3] = np.cos(theta)
        obs[4] = theta_dot
        return obs



from copy import deepcopy
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F
from typing import Tuple
from .base import OneStepAgent
from ..config import Config
from ..prelude import Action, Array, State


class DDPGAgent(OneStepAgent):
    SAVED_MEMBERS = "net", "target_net", "actor_opt", "critic_opt", "explorer", "replay"

    def __init__(self, config: Config) -> None:
        super().__init__(config)
        self.net = config.net("ddpg")
        self.target_net = deepcopy(self.net)
        self.actor_opt = config.optimizer(self.net.actor_params(), key="actor")
        self.critic_opt = config.optimizer(self.net.critic_params(), key="critic")
        self.explorer = config.explorer()
        self.eval_explorer = config.explorer(key="eval")
        self.replay = config.replay_buffer()
        self.batch_indices = config.device.indices(config.replay_batch_size)

    def set_mode(self, train: bool = True) -> None:
        self.net.train(mode=train)

    @torch.no_grad()
    def eval_action(self, state: Array) -> Action:
        action = self.eval_explorer.add_noise(self.net.action(state))
        return action.cpu().numpy()  # type: ignore

    def step(self, state: State) -> Tuple[State, float, bool, dict]:
        train_started = self.total_steps > self.config.train_start
        if train_started:
            with torch.no_grad():
                action = self.net.action(state)
            action = self.explorer.add_noise(action).cpu().numpy()
        else:
            action = self.env.spec.random_action()
        action = self.env.spec.clip_action(action)
        next_state, reward, done, info = self.env.step(action)
        self.replay.append(state, action, reward, next_state, done)
        if train_started:
            self._train()
        return next_state, reward, done, info

    @torch.no_grad()
    def _q_next(self, next_states: Array) -> Tensor:
        actions = self.target_net.action(next_states)
        return self.target_net.q_value(next_states, actions)

    def _train(self) -> None:
        obs = self.replay.sample(self.config.replay_batch_size)
        obs = [ob.to_array(self.env.extract) for ob in obs]
        states, actions, rewards, next_states, done = map(np.asarray, zip(*obs))
        mask = self.config.device.tensor(1.0 - done)
        q_next = self._q_next(next_states)
        q_target = (
            q_next.squeeze_()
            .mul_(mask * self.config.discount_factor)
            .add_(self.config.device.tensor(rewards))
        )
        action, q_current = self.net(states, actions)

        #  Backward critic loss
        critic_loss = F.mse_loss(q_current.squeeze_(), q_target)
        self._backward(critic_loss, self.critic_opt, self.net.critic_params())

        #  Backward policy loss
        policy_loss = -self.net.q_value(states, action).mean()
        self._backward(policy_loss, self.actor_opt, self.net.actor_params())

        #  Update target network
        self.target_net.soft_update(self.net, self.config.soft_update_coef)



from abc import ABC, abstractmethod
from itertools import chain
from rainy.utils import Device
import torch
from torch import nn, Tensor
from typing import Iterable, List, Sequence, Tuple
from .block import FcBody, LinearHead, NetworkBlock
from .init import Initializer, kaiming_uniform
from .misc import SoftUpdate
from .prelude import NetFn
from .value import ContinuousQFunction
from ..prelude import ArrayLike


class DeterministicPolicyNet(ABC):
    @abstractmethod
    def action(self, state: ArrayLike) -> Tensor:
        pass


class DDPGNet(SoftUpdate, ContinuousQFunction, DeterministicPolicyNet):
    pass


class SeparatedDDPGNet(DDPGNet):
    def __init__(
        self,
        actor_body: NetworkBlock,
        critic_body: NetworkBlock,
        action_dim: int,
        action_coef: float = 1.0,
        device: Device = Device(),
        init: Initializer = Initializer(weight_init=kaiming_uniform(a=3 ** 0.5)),
    ) -> None:
        super().__init__()
        self.actor = nn.Sequential(
            actor_body,
            LinearHead(actor_body.output_dim, action_dim, init=init),
            nn.Tanh(),
        )
        self.critic = nn.Sequential(
            critic_body, LinearHead(critic_body.output_dim, 1, init=init)
        )
        self.to(device.unwrapped)
        self.action_coef = action_coef
        self.device = device

    def actor_params(self) -> Iterable[Tensor]:
        return self.actor.parameters()

    def critic_params(self) -> Iterable[Tensor]:
        return self.critic.parameters()

    def action(self, states: ArrayLike) -> Tensor:
        s = self.device.tensor(states)
        return self.actor(s).mul(self.action_coef)

    def q_value(self, states: ArrayLike, action: ArrayLike) -> Tensor:
        s = self.device.tensor(states)
        a = self.device.tensor(action)
        sa = torch.cat((s, a), dim=1)
        return self.critic(sa)

    def forward(self, states: ArrayLike, action: ArrayLike) -> Tuple[Tensor, Tensor]:
        s = self.device.tensor(states)
        a = self.device.tensor(action)
        sa = torch.cat((s, a), dim=1)
        return self.actor(s).mul(self.action_coef), self.critic(sa)


class SeparatedTD3Net(SeparatedDDPGNet):
    def __init__(
        self,
        actor_body: NetworkBlock,
        critic_body: NetworkBlock,
        critic_body2: NetworkBlock,
        action_dim: int,
        action_coef: float = 1.0,
        device: Device = Device(),
        init: Initializer = Initializer(weight_init=kaiming_uniform(a=3 ** 0.5)),
    ) -> None:
        super().__init__(
            actor_body,
            critic_body,
            action_dim,
            action_coef=action_coef,
            device=device,
            init=init,
        )
        self.critic2 = nn.Sequential(
            critic_body2, LinearHead(critic_body2.output_dim, 1, init=init)
        )
        self.to(device.unwrapped)

    def critic_params(self) -> Iterable[Tensor]:
        return chain(self.critic.parameters(), self.critic2.parameters())

    def q_value(self, states: ArrayLike, action: ArrayLike) -> Tensor:
        s = self.device.tensor(states)
        a = self.device.tensor(action)
        sa = torch.cat((s, a), dim=1)
        return self.critic(sa), self.critic2(sa)


def fc_seprated(
    action_coef: float = 1.0,
    actor_units: List[int] = [400, 300],
    critic_units: List[int] = [400, 300],
    init: Initializer = Initializer(weight_init=kaiming_uniform(a=3 ** 0.5)),
) -> NetFn:
    """DDPG network with separated bodys
    """

    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> SeparatedDDPGNet:
        actor_body = FcBody(state_dim[0], units=actor_units, init=init)
        critic_body = FcBody(state_dim[0] + action_dim, units=critic_units, init=init)
        return SeparatedDDPGNet(
            actor_body,
            critic_body,
            action_dim,
            action_coef=action_coef,
            device=device,
            init=init,
        )

    return _net


def td3_fc_seprated(
    action_coef: float = 1.0,
    actor_units: List[int] = [400, 300],
    critic_units: List[int] = [400, 300],
    init: Initializer = Initializer(weight_init=kaiming_uniform(a=3 ** 0.5)),
) -> NetFn:
    """TD3 network with separated bodys
    """

    def _net(
        state_dim: Sequence[int], action_dim: int, device: Device
    ) -> SeparatedTD3Net:
        actor_body = FcBody(state_dim[0], units=actor_units, init=init)
        critic1 = FcBody(state_dim[0] + action_dim, units=critic_units, init=init)
        critic2 = FcBody(state_dim[0] + action_dim, units=critic_units, init=init)
        return SeparatedTD3Net(
            actor_body,
            critic1,
            critic2,
            action_dim,
            action_coef=action_coef,
            device=device,
            init=init,
        )

    return _net
